"""Example demonstrating a basic usage of choke package."""
from time import sleep
from redis import StrictRedis
from choke import RedisChokeManager, CallLimitExceededError


REDIS = StrictRedis() # Tweak this to reflect your setup
CHOKE_MANAGER = RedisChokeManager(redis=REDIS)

# Example configuration: enforce limit of no more than 10 calls in two seconds window
@CHOKE_MANAGER.choke(limit=10, window_length=2)
def foo(x, y):
    """Just print something to show that foo was called."""
    print(f'foo called with ({x}, {y})')

if __name__ == '__main__':
    # We expect pattern of 10 successes followed by 10 failures followed again by 10 successes
    # Some deviations from this pattern may obviously occur as calling foo takes nonzero time
    for i in range(30):
        try:
            foo(i, y=i ** 2)
        except CallLimitExceededError:
            print('Foo not called. Limit exceeded!')
        sleep(0.1)



"""Implementation of simple choke based on dictionary of lists."""
import bisect
from collections import defaultdict
from time import time
from choke.choke_base import BaseChokeManager


class SimpleChokeManager(BaseChokeManager):
    """Simplest, non-thread safe choke utilizing dict mapping tags to lists of timestamps."""

    def get_records(self, tag):
        """Get records corresponding to given tag."""
        return tuple(self.records[tag])

    def __init__(self, time_source=time):
        super(SimpleChokeManager, self).__init__(time_source)
        self.records = defaultdict(list)

    def _register_timestamp(self, tag, timestamp):
        bisect.insort(self.records[tag], timestamp)

    def _ltrim_timestamps(self, tag, max_timestamp):
        index = bisect.bisect_right(self.records[tag], max_timestamp)
        self.records[tag] = self.records[tag][index:]

    def _count_timestamps_for_window(self, tag, timestamp, window_length):
        left = bisect.bisect_right(self.records[tag], timestamp - window_length)
        right = bisect.bisect_right(self.records[tag], timestamp)
        return right - left



"""Test cases for redis_choke module."""
from itertools import chain, repeat
import pytest
from choke.redis_choke import RedisChokeManager


def test_calls_zadd(mocker):
    """RedisChokeManager should call zadd of its redis object when registering timestamp."""
    redis_mock = mocker.Mock()
    time_mock = mocker.Mock(return_value=100.0)
    choke = RedisChokeManager(redis_mock, time_source=time_mock)

    choke.register_timestamp('test_tag')

    redis_mock.zadd.assert_called_once_with('test_tag', {100.0: 100.0})

def test_calls_zremrangebyscore(mocker):
    """RedisChokeManager should call redis.zremrangebyscore when pruning timestamps."""
    redis_mock = mocker.Mock()
    time_mock = mocker.Mock(return_value=100)
    choke = RedisChokeManager(redis_mock, time_source=time_mock)

    choke.prune('some_tag', 1000, ref_timestamp=4120.0)

    redis_mock.zremrangebyscore.assert_called_once_with('some_tag', 0, 3120.0)

def test_obtains_reference_timestamp(mocker):
    """RedisChokeManager should use its time_source to obtain reference timestamp when pruning."""
    redis_mock = mocker.Mock()
    time_mock = mocker.Mock(return_value=2137.0)
    choke = RedisChokeManager(redis_mock, time_source=time_mock)

    choke.prune('some_tag', 1000)
    time_mock.assert_called_once_with()
    redis_mock.zremrangebyscore.assert_called_once_with('some_tag', 0, 1137.0)

def test_calls_zcount(mocker):
    """RedisChokeManager should call redis.zcount for obtaining count of timestamps."""
    redis_mock = mocker.Mock()
    redis_mock.zcount.return_value = 31
    time_mock = mocker.Mock(return_value=6415.0)
    choke = RedisChokeManager(redis_mock, time_source=time_mock)

    assert choke.count_records('tag123', 27) == 31
    redis_mock.zcount.assert_called_once_with('tag123', '(6388.0', 6415.0)



"""Initialization file for choke package."""
from choke.redis_choke import RedisChokeManager
from choke.simple_choke import SimpleChokeManager
from choke.choke_base import CallLimitExceededError

__all__ = ['RedisChokeManager', 'SimpleChokeManager', 'CallLimitExceededError']



"""Test cases for simple_choke module."""
from itertools import chain, repeat
import pytest
from choke.simple_choke import SimpleChokeManager


def test_registers_timestamp(mocker):
    """SimpleChokeManager should correctly register timestamps."""
    time = mocker.Mock(side_effect=[10, 100, 190])
    choke_manager = SimpleChokeManager(time_source=time)

    for _ in range(3):
        choke_manager.register_timestamp('mytag')

    assert choke_manager.get_records('mytag') == (10, 100, 190)

def test_prune(mocker):
    """SimpleChokeManager should prune records using current time and provided window_length."""
    time = mocker.Mock(side_effect=[100, 250, 300, 350, 400, 450, 450])
    choke_manager = SimpleChokeManager(time_source=time)

    for _ in range(4):
        choke_manager.register_timestamp('mytag')

    choke_manager.prune('mytag', window_length=200)
    assert choke_manager.get_records('mytag') == (250, 300, 350)
    choke_manager.prune('mytag', window_length=200)
    assert choke_manager.get_records('mytag') == (300, 350)
    choke_manager.prune('mytag', window_length=150)
    assert choke_manager.get_records('mytag') == (350,)

@pytest.mark.parametrize(
    'timestamps, window_length, exp_count',
    [
        [[21, 37, 39, 43], 80, 3],
        [[21, 37, 39, 43], 7, 2],
        [[21, 37, 39, 43], 6, 1],
        [[21, 37, 39, 43], 3, 0]
    ])
def test_count_noprune(timestamps, window_length, exp_count, mocker):
    """SimpleChokeManager should correctly count registered timestamps wnen prune=False."""
    time = mocker.Mock(side_effect=timestamps)
    choke_manager = SimpleChokeManager(time_source=time)

    for _ in range(len(timestamps)-1):
        choke_manager.register_timestamp('mytag')

    assert choke_manager.count_records('mytag', window_length, prune=False) == exp_count

def test_count_prune(mocker):
    """SimpleChokeManager should correctly count registered timestamps wnen prune=True."""
    time = mocker.Mock(side_effect=chain([100, 200, 300, 400], repeat(401)))
    choke_manager = SimpleChokeManager(time_source=time)

    for _ in range(4):
        choke_manager.register_timestamp('mytag')

    assert choke_manager.count_records('mytag', window_length=400) == 4
    assert choke_manager.count_records('mytag', window_length=200) == 2

    # After pruning this value should change
    assert choke_manager.count_records('mytag', window_length=400) == 2



"""Redis-based implementation of throttling mechanism."""
from time import time
from choke.choke_base import BaseChokeManager


class RedisChokeManager(BaseChokeManager):
    """Basic class implementing low-level logic for choking resources."""

    def __init__(self, redis, time_source=time):
        super(RedisChokeManager, self).__init__(time_source)
        self.redis = redis

    def _register_timestamp(self, tag, timestamp):
        self.redis.zadd(tag, {timestamp: timestamp})

    def _ltrim_timestamps(self, tag, max_timestamp):
        self.redis.zremrangebyscore(tag, 0, max_timestamp)

    def _count_timestamps_for_window(self, tag, timestamp, window_length):
        return self.redis.zcount(tag, f'({timestamp - window_length}', timestamp)



"""choke: simple implementation of throttling mechanism."""

from setuptools import setup, find_packages

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Software Development']

with open('README.rst') as readme:
    LONG_DESCRIPTION = readme.read()

setup(
    name='choke',
    license='MIT',
    description=__doc__,
    use_scm_version=True,
    long_description=LONG_DESCRIPTION,
    platforms=["Linux", "Unix"],
    setup_requires=['setuptools_scm'],
    install_requires=['redis'],
    tests_require=['pytest', 'pytest-mock'],
    author='Konrad Jałowiecki <dexter2206@gmail.com>',
    author_email='dexter2206@gmail.com',
    packages=find_packages(exclude=["tests", "tests.*"]),
    keywords='throttling redis limit'
)



"""Common functionallities used across choke package."""
import abc
from functools import wraps
import logging
import time


class CallLimitExceededError(RuntimeError):
    """Exception raised when call limit is exceeded for a choked action."""


class BaseChokeManager(abc.ABC):
    """ABC for other choke managers."""

    def __init__(self, time_source=time):
        self.time_source = time_source

    def count_records(self, tag, window_length, prune=True):
        """Get count of records for given tag in given window, optionally pruning old records."""
        timestamp = self.time_source()
        logger = self.get_logger(tag)

        if prune:
            self.prune(tag, window_length, timestamp)

        logger.debug('Checking cardinality of timestamps set for tag %s in recent window.', tag)
        count = self._count_timestamps_for_window(tag, timestamp, window_length)
        logger.debug('Number of records for tag %s in last window of length %s: %d',
                     tag, window_length, count)

        return count

    def prune(self, tag, window_length, ref_timestamp=None):
        """Prune timestamps for the given tag outside of given window."""
        logger = self.get_logger(tag)
        ref_timestamp = ref_timestamp or self.time_source()
        max_timestamp = ref_timestamp - window_length

        logger.debug('Pruning timestamps below %s for tag %s.', max_timestamp, tag)
        self._ltrim_timestamps(tag, max_timestamp)

    def choke(self, window_length, limit, name=None):
        """Create a factory for producing "choked" callables.

        The choked callables implement automated registering timestamps when they are called
        and enforcing limits placed on the number of calls in defined time window.
        """
        def _choked_action_factory(target):
            tag = name or target.__name__
            logger = self.get_logger(tag)
            @wraps(target)
            def _choked_action(*args, **kwargs):
                logger.debug("Getting records' count.")
                count = self.count_records(tag, window_length, prune=True)
                if count >= limit:
                    logger.debug('Call limit exceeded. Callable choked.')
                    raise CallLimitExceededError(f'Limit exceeded for callable {tag}.')
                logger.debug('Limit not exceeded (%d/%d calls in current time window)', count, limit)
                self.register_timestamp(tag)
                logger.debug('Forwarding call to wrapped callable.')
                return target(*args, **kwargs)
            return _choked_action
        return _choked_action_factory

    def register_timestamp(self, tag):
        """Register new timestamp for given tag and window."""
        timestamp = self.time_source()
        logger = self.get_logger(tag)

        self._register_timestamp(tag, timestamp)
        logger.debug('New timestamp %s registered for tag %s.', timestamp, tag)

    @staticmethod
    def get_logger(tag):
        return logging.getLogger(f'redis_choke.{tag}')

    @abc.abstractmethod
    def _count_timestamps_for_window(self, tag, timestamp, window_length):
        pass

    @abc.abstractmethod
    def _ltrim_timestamps(self, tag, max_timestamp):
        pass

    @abc.abstractmethod
    def _register_timestamp(self, tag, timestamp):
        pass
